"""
pauli_utils.py
@author Ben Reggio

This class contains methods which fall into miscallaneous categories, which are used by the classes:
matrix_generator.py
Psfam.py
family.py

These methods include methods which are self-explanatory, or methods distract from the main purpose of these classes
This code will show how these methods are made if one is looking for more detail.

The following methods are contained in this class:

transpose
invert
mtimes
is_invertible
get_coords
increment_coords
increment_string
get_progression
m_times
mat_times_vec
string_process
tobin
get_string
bin_add
to_int
enum_strings
"""


import galois as gf
GF2 = gf.GF(2)          #Some vectors and matrices will be in GF(2)


def dotp(a,b):
    m = len(a)
    s=0
    for i in range(m):
        s = s + a[i]*b[i]
    return s

#Util method to get transpose of a matrix
def transpose(M):
  M2 = []
  for i in range(len(M)):
    row = []
    for j in range(len(M)):
      row = row + [M[j][i]]
    M2 = M2 + [row]
  return M2

#Util method to invert a square GF(2) valued matrix
def invert(C):
  m = len(C)
  A=[]
  for i in range(m):                        #Construct augmented matrix
    row = [GF2(0)]*m
    row[i] = GF2(1)
    A = A + [C[i] + row]
  
  for i in range(m):                        #build upper triangular matrix
    j=i+1
    while(A[i][i] != GF2(1)):               #Swap rows such that the diagonal is 1
      if A[j][i] == GF2(1):
        t = A[j].copy()
        A[j] = A[i].copy()
        A[i] = t
      j=j+1

    for j in range(i+1,m):                  #Add rows below such that the bottom triangular area is 0
      if A[j][i] == GF2(1):
        for k in range(2*m):
          A[j][k] = A[i][k] + A[j][k]

  for i in range(m):                        #Clear upper triangular area to 0
    for j in range(i+1,m):
      if A[i][j] == 1:
        for k in range(2*m):
          A[i][k] = A[i][k] + A[j][k]
  
  B=[]
  for i in range(m):
    B = B + [A[i][m:]]                      #Select augmented side
  return B

#Util method to multiply matrices or vectors
def mtimes(M1,M2):
  if(type(M2[0]) == list):              #If matrix times matrix:
    M3 = []
    m = len(M1)
    for i in range(m):
      row = []
      for j in range(m):
        s=GF2(0)
        for k in range(m):              # sum M1_{i,k} M2_{k,j}
          s = s + M1[i][k]*M2[k][j]
        row = row + [s]
      M3 = M3 + [row]
    return M3
  else:                                 #If matrix times vector:
    r = []
    m = len(M1)
    for j in range(m):
      s=GF2(0)
      for k in range(m):
          s = s + M1[j][k]*M2[k]        #sum M1_{i,k} V_{k}
      r = r + [s]
    return r

#Util method to determine if a matrix is invertible. Goes through part of the process of inverting a matrix
def is_invertible(B):
  m = len(B)
  A = []
  for i in range(m):
    row = [GF2(0)]*m
    row[i] = GF2(1)
    A = A + [B[i].copy() + row]
  
  for i in range(m):
    j=i+1
    while(A[i][i] != GF2(1)):
      if j>=m:
        return False
      if A[j][i] == GF2(1):
        t = A[j].copy()
        A[j] = A[i].copy()
        A[i] = t
      j=j+1


    for l in range(i+1,m):
      if A[l][i] == GF2(1):
        for k in range(2*m):
          A[l][k] = A[i][k] + A[l][k]

  return True

#Utility method which gets the coordinates in the X-Z table of the pauli string. For example, XIY = XIX * IIZ = 5,1
def get_coords(pauli_input,m="unspecified"):
    ps = pauli_input
    if(type(ps) == int):
        ps=list(str(ps))
    if(type(ps) == str):
        ps = list(ps)
    i = []
    j = []
    if(type(ps) == list):
        if(m == "unspecified"):
            m = len(ps)
        else:
            if m != len(ps):
                print("ERROR: Wrong pauli size. m = " + str(m) + " the string " + str(pauli_input) + " has size " + str(len(ps)))
                return null
        ps = list(ps)

        if ps[0] in str_chars:
            for k in range(m):
                if(ps[k] == 'I') or (ps[k] == 'X'):
                    i = i + [0]
                else:
                    i = i + [1]
            for k in range(m):
                if(ps[k] == 'I') or (ps[k] == 'Z'):
                    j = j + [0]
                else:
                    j = j + [1]
        if ps[0] in int_chars:
            for k in range(m):
                if(ps[k] == '0') or (ps[k] == '2'):
                    i = i + [0]
                else:
                    i = i + [1]
            for k in range(m):
                if(ps[k] == '0') or (ps[k] == '1'):
                    j = j + [0]
                else:
                    j = j + [1]
    while(len(i)<m):
        i = [0] + i
    while(len(j)<m):
        j = [0] + j
    return to_int(i),to_int(j)

#Utility method to generate all strings from the coordinates of the X-Z table
tuple_progression = [(0,0),(1,0),(0,1),(1,1)]
def increment_coords(list_tuple,n):
    new_z,new_x = list_tuple
    ct = (new_z[n],new_x[n])
    for i in range(3):
        if ct == tuple_progression[i]:
            ct_z,ct_x = tuple_progression[i+1]
            new_x[n] = ct_x
            new_z[n] = ct_z
            return (new_z,new_x)
    ct_z,ct_x = tuple_progression[0]
    new_x[n] = ct_x
    new_z[n] = ct_z
    return increment_coords((new_z,new_x),n-1)

#Utility method to increment a string along a table. III -> IIZ -> IIX -> IIY -> IZI etc.
#This can be used to get every string in a specific order. n should be the size of the string.
def increment_string(s,n):
    new_s = s
    for i in range(3):
        #print("s = " + str(s))
        if s[n] == str_chars[i]:
            new_s[n] = str_chars[i+1]
            return new_s
    new_s[n] = "I"
    return increment_string(s,n-1)
    
    
int_chars = ['0','1','2','3']
#Utility method to multiply a matrix A with a vector v in Z2
def mat_times_vec(A,v):
    m = len(A)
    r = []
    for i in range(m):
        s=0
        for j in range(m):
            s = s + A[i][j] * v[j]
        r = r + [s%2]
    return r

str_chars=['I','Z','X','Y']
def string_process(s):
  ns = ''
  for i in range(len(s)):
    ns = ns + str_chars[(int)(s[i])]
  return ns

#Utility method to get the binary representation of an integer with length s.
def tobin(x,s):
    return [(x>>k)&1 for k in range(s-1,-1,-1)]

#Utility method to get the string given its coordinates on the X-Z table
def get_string(i,j):
  c = [str_chars[i[l] + 2*j[l]] for l in range(len(i))]
  return "".join(c)

#Utility method to get the bitwise sum of two integers (3 + 5 = 6)
def bin_add(a,b):
  c=[]
  for i in range(len(a)):
    if(a[i] == b[i]):
      c = c + [0]
    else:
      c = c + [1]
  return c

#Utility method for representing a binary vector as an integer
def to_int(a):
  m=len(a)
  s=0
  for j in range(m):
    if(a[m-j-1] == 1):
      s = s + 2**j
  return s

#Utility method to enumerate all of the pauli strings of length m.
def enum_strings(m):
    s1 = ['I']*m
    sl=[]
    for i in range(4**m):
        sl = sl + ["".join(s1)]
        increment_string(s1,m-1)
    return sl

#This is used for printing out a GF2 valued matrix in a readable manner
def show_GF_mat(mat):
    for row in mat:
        newrow = []
        for i in range(len(row)):
            if row[i] == GF2(0):
                newrow = newrow + [0]
            else:
                newrow = newrow + [1]
        print(newrow)


'''
Psfam.py
@author: Ben Reggio

The purpose of this code is to create an object Pauli_Organizer, which assists VQE by partitioning pauli strings into the minimum set of commuting families. The following things are required.

1 parameter m - representing the size of the pauli strings (number of characters in the string). This should be given by the user based on the size of the 

A symmetric matrix which is similar to a companion matrix. This is solved the class matrix_generator, which has a method get_generator_matrix(m) which takes m and returns the appropriate A. A^i gives the matrices which represent the different families

The object Pauli_Organizer contains family objects which have the information about these individual families.
They will be supplied with the following information from Pauli_organizer:
    - The size m of the strings
    - The list representation of the family
    - A representation of the correct diagonalization of the family through the set_q(q) method
    - The correct coefficients of each measurement if such a decomposition is fed to this object through the input_pauli_decomps() method

A user may call get_families() method to get a list of the 2**m + 1 family objects
'''


from psfam.matrix_generator import *        #To get A matrix which is symmetric and similar to a companion matrix
from psfam.family import *                                  #Contains the important family class

from psfam.pauli_utils import *      #Self-explanatory methods or methods which are not relevant to the main purpose of
                               #Pauli_organizer are stored in pauli_utils
    
#Main user focused class. This class will take an integer m as an input and generate a set of family objects
#The family objects will contain a perfect partition of the strings.
#
#A secondary function of this class is to assign the correct circuits to the families. While generating the families,
#this object uses the set_q function of the families to assign the correct post-circuit rotations to the family objects
#
#A third functionality is to take in a decomposition of a Hamiltonian into Pauli Strings, and output the correct
#coefficients of each measurement so that the VQE circuit can be evaluated in the minimum amout of steps
class PauliOrganizer:

    #Constructor initiates partitioning:
    def __init__(self,m):
        super().__init__()
        self.m = m
        self.u = False
        self.N = 2**m
        self.f = self.build_family_group()                              #Partition strings
        self.pauli_decomposition = self.initialize_decomposition()      #set pauli_decomposition to 0 so that it can
                                                                        #it can be set individually if necessary
    
        
    #Returns list of families
    def get_families(self):
        return self.f

    #Used for setting the entirety of a Hamilton decomposition into pauli_decomposition parameter
    def input_pauli_decomps(self,pauli_dict):
        coords = ([0]*self.m,[0]*self.m)                                    #coords gives the coordinates of cp
        cp = ['I']*self.m                                                   #cp (current pauli) begins as the identity
        self.pauli_decomposition[0][0] = pauli_dict["".join(cp)]            #inputs the coefficient of cp in the input pauli_dict
        for k in range(self.N ** 2):
            #print("coords = " + str(coords))
            #print('cp = ' + str("".join(cp)))
            coords = increment_coords(coords,self.m - 1)                    #Increase coords to next Z_2 value
            cp = increment_string(cp,self.m - 1)                            #Go to next string with next coords
            i = to_int(coords[0])
            j = to_int(coords[1])
            self.pauli_decomposition[i][j] = pauli_dict["".join(cp)]        #inputs next string decomposition into
                                                                            #pauli_decomposition
        
    #Use this to input the decomposition string by string
    def input_pauli_decomp(self,pauli_string,value):
        i,j = get_coords(pauli_string,self.m)
        self.pauli_decomposition[i][j] = value
    
    #Sets decomposition to 0 so that it can be entered individually
    def initialize_decomposition(self):
        pc = []
        for i in range(self.N):
            row = []
            for j in range(self.N):
                row = row + [0]
            pc = pc + [row]
        return pc
     
    #This function partitions the strings and creates the families.
    def build_family_group(self):
        A = get_generating_matrix(self.m)
        self.generating_matrix = A
        sol = process_A(A,self.m)
        f=[]
        for i in range(self.N - 1):
            fam = family(sol[i],self.m)
            q_i = 0
            if(i%2 == 1):
                q_i = (int)((i+1)/2) - 1
            else:
                q_i = (int)((self.N + i)/2) - 1
            fam.set_q(sol[q_i])
            f = f + [fam]
        f = f + [xfamily(self.m),zfamily(self.m)]
        return f
    
    #Print the properties of this solution
    def properties(self):
        s = f''
        s = s + 'Qubits: ' + str(self.m) +  '\nGenerating Matrix:'
        A = self.generating_matrix
        for row in A:
            s = s + '\n' + str(row)
        return s
    
    #Uses pauli_decomposition to calculate the coefficients of each measurement
    def calc_coefficients(self):
        f=self.f
        N = len(f) - 1
        for family_index in range(N - 1):
            p = f[family_index].p
            c=[]
            for state_index in range(0,N):
                s = 0
                for string_index in range(1,N):
                    v = eigenvalue(string_index,state_index)
                    effect = self.pauli_decomposition[string_index][p[string_index-1]]*v*f[family_index].signs[string_index-1]
                    s = s + effect
                c=c+[s]
            f[family_index].set_coefficients(c)
        x_index = N-1
        z_index = N
        c=[]
        for state_index in range(0,N):
            s = 0
            for string_index in range(1,N):
                v = eigenvalue(string_index,state_index)
                effect = self.pauli_decomposition[0][string_index]*v
                s = s + effect
            c=c+[s]
        f[x_index].set_coefficients(c)
        c=[]
        for state_index in range(0,N):
            s = 0
            for string_index in range(0,N):
                v = eigenvalue(string_index,state_index)
                effect = self.pauli_decomposition[string_index][0]*v
                s = s + effect
            c=c+[s]
        f[z_index].set_coefficients(c)
        return f

    def get_family_of_string(self,s):
        p = self.f[0].p
        a = 1
        u=[]
        if(self.u == False):
            for l in range(self.N -1):
                u = u + [a]
                a = p[a-1]
            self.u = u
        else:
            u = self.u
        i,j = util.get_coords(s)
        if(j == 0):
            return self.f[self.N]
        if(i == 0):
            return self.f[self.N -1]
        k = (u.index(j) - u.index(i) - 1) % (self.N -1)
        return self.f[k]

    def get_fam_index(self,s):
        p = self.f[0].p
        a = 1
        u=[]
        if(self.u == False):
            for l in range(self.N -1):
                u = u + [a]
                a = p[a-1]
            self.u = u
        else:
            u = self.u
        i,j = util.get_coords(s)
        if(j == 0):
            return self.N
        if(i == 0):
            return self.N - 1
        k = (u.index(j) - u.index(i) - 1) % (self.N -1)
        return k


#This function takes the generating matrix as an input, and outputs a list of each family in its permutation representation.
#For example, the one element in the list might be [2,3,1] representing the family XZ YX ZY.
def process_A(A,m):
  l1 = []                               #List representation of family 1
  for i in range(m):                    #Each ROW in A represents a generating string in the family
    t = A[m-i-1]                        #t represents the generating string
    l1 = l1 + [t]
    
    #Here, we use linearity of the solution and carefully order the elements in our list. l1[i] is the permutation P(i). Therefore, once we know l[i] and l[j], we know l[i + j]. Every time we add another element to the list, we can add it to each previous element.
    for j in range(len(l1)-1):
      l1 = l1 + [bin_add(l1[j],t)]
    
  for i in range(len(l1)):
    l1[i] = to_int(l1[i])                       #Reformat binary vectors as integers
  
  data = [l1]                                   #data will contain the entire output
  prev_p = l1.copy()                            #prev_p is the permutation of the previous family
  N = 2**m - 1
  for i in range(N-1):
    next_p = []                                 #next_p is the permutation of the next family (the one we are making)
    for j in range(N):
      next_p = next_p + [prev_p[l1[j]-1]]       #This statement is equivalent to P^(i+1)(x) = P(P^(i)(x)) 
    data = data + [next_p]
    prev_p = next_p.copy()
  return data

#Method to get the dot product of v_i and v_j which represent the vectors in z_2^m
#So eigenvalue(5,6) can be found by doing 5 -> (1,0,1) ; 6 -> (1, 1, 0). (1,0,1) dot (1,1,0) = 1 + 0 + 0 = 1 
def eigenvalue(i,j):
    ib = list(bin(i)[2:])
    jb = list(bin(j)[2:])
    a = len(ib)
    b = len(jb)
    m=max(a,b)
    ib = tobin(i,m)
    jb = tobin(j,m)
    s=0
    for k in range(m):
        if((ib[k] == 1) & (jb[k] == 1)):
            s = s + 1
    s = (s+1)%2
    if s == 0:
        s = -1
    return s


'''
family.py
@author: Ben Reggio

The class family is used to hold information about commuting families of pauli strings. Families have three
representations. The first representation is a string representation: IY YI YY. The second is a list representation,
where a list p[i] represents the strings x_i z_p[i] or z_i x_p[i]. This would be [1,2,3]. The final representation
is a matrix represnetation, where the matrix acting on the binary representation of i gives p[i]. This would be:
A = [1,0]
    [0,1]
Since A v_1 = v_1; and Av_2 = v_2. This returns the correct p[i]. This family should have access to all three
representations. The list representation will be used in the constructor. This class should be able to get all
other representations from this one.

The other functionality of this class is that it should be able to diagonalize the family. A family with a Matrix
representation A has a corresponding A^{N/2 % N-1}. When this is entered using set_q(q), the family object
will be able to diagonalize itself. The family will take the form U D U dagger, where D represents a diagonal
matrix which is equal to a pauli string in the z family. In this case U = exp(sum(x_i)) where x_i will be the
x strings in the x_list parameter, which is filled when the set(q) method is called.

After diagonalization, this code integrates with qiskit with the method apply_to_circuit(circuit). It will
add onto the end of the circuit the clifford gate corresponding to the matrix U.
'''

import psfam.pauli_utils as util      #methods tobin and get_string are used from pauli_utils
from numpy import pi

#This class will contain the information for each family of commmuting Pauli strings. It should be able to give the 
#String representation of the family with to_string(), the list representation with get_permutation(), and the matrix
#representation using get_generating_matrix().
class family:
    
    #Constructor takes in the list representation of the family and the size of the strings.
    def __init__(self,p,m):
        self.p = p
        self.q = False
        self.m = m
        self.coefficients = "unassigned"
        self.signs = [0]*(2**m -1)
        
    #q is used to determine the post-rotation unitary matrix that diagonalizes this family. 
    #q should be the N/2 or 1/2 power of the generating matrix in permutation form. The generators of
    #this permutation are placed in x_list, so that U = exp(sum x_i) where x_i are the strings in x_list
    def set_q(self,q):
        self.q = q
        l=[]
        for i in range(self.m):
            l = l + [q[2**i-1]]
        self.x_list=l
        for i in range(1,2**self.m):
            self.get_sign(i)
        
    #Used to print out all of the strings in this family in string form
    def __str__(self):
        p = self.p
        ps = []
        m = self.m
        for i in range(len(p)):
            bj = util.tobin(p[i],m)
            bi = util.tobin(i+1,m)
            ps = ps + [util.get_string(bi,bj)]
        return ",".join(ps)

    #Used to print out all of the strings in this family in string form
    def to_string(self):
        p = self.p
        ps = []
        m = self.m
        for i in range(len(p)):
            bj = util.tobin(p[i],m)
            bi = util.tobin(i+1,m)
            ps = ps + [util.get_string(bi,bj)]
        return ps
    
    #Given a quantum circuit object circuit, this applies the correct rotation for U which diagonalizes the family
    def apply_to_circuit(self,circuit):
        apply_e_x_series(circuit,self.x_list,self.m)
    
    #Set the coefficients of each measurement
    def set_coefficients(self,c):
        self.coefficients = c
      
    #Get the coefficients of each measurement
    def get_coefficients(self):
        return self.coefficients
    
    def get_permutation(self):
        return self.p
    
    #Return the m x m symmetric matrix representing this family
    def get_generating_matrix(self):
        data = []
        for i in range(self.m):
            row = util.tobin(self.p[2**i - 1],self.m)
            data = [row] + data
        return data

    #Summary of this family
    def properties(self):
        s = f''
        s = s + 'Qubits: ' + str(self.m) + '\nPermutation:' + str(self.get_permutation())
        s = s + '\nCoefficients: ' + str(self.get_coefficients()) + '\nGenerating Matrix:'
        A = self.get_generating_matrix()
        for row in A:
            s = s + '\n' + str(row)
        return s

    #Print strings used to diagonalize family. The unitary U = exp(i pi/4 (sum P_i)) is used, where P_i are
    #the strings which result from this method
    def get_diagonalizing_strings(self):
        ds = []
        for i in range(self.m):
            j = 2**i - 1
            ds = ds + [util.get_string(util.tobin(0,self.m),util.tobin(self.q[j],self.m))]
        return ds

    def get_sign(self,i):
        pb = util.tobin(i,self.m)
        #print(self.to_string()[i-1])
        #print("S_ " + str(i) + " " + str(self.p[i-1]))
        totx = [0]*self.m
        s=0
        for j in self.x_list:
            jb = util.tobin(j,self.m)
            dtp = util.dotp(pb,jb)
            if(dtp%2 == 1):
                #totx = util.bin_add(totx,jb)
                s=s+1
        #print(util.dotp(pb,totx))
        #print(s)
        #print(totx)
        #print(util.tobin(self.p[i-1],self.m))
        if((util.dotp(pb,util.tobin(self.p[i-1],self.m)) + s) % 4 == 0):
            self.signs[i-1] = 1
        else:
            self.signs[i-1] = -1

#Extra classes for the x and z families. Should behave very similar to family, but won't break for not having
#a permutation that goes along with it.
class xfamily(family):
    def __init__(self,m):
        self.p = False
        self.q = False
        self.m = m
        self.coefficients = "unassigned"
        self.signs = [1]*(2**self.m - 1)

    def get_diagonalizing_strings(self):
        ds = []
        for i in range(self.m):
            s = ["I"]*self.m 
            s[i] = "Y"
            ds = ds + ["".join(s)]
        return ds
        
    def get_permutation(self):
        return False
    
    def apply_to_circuit(self,circuit):
        for i in range(self.m):
            circuit.u(-pi/2,0,0,i)
    
    def __str__(self):
        p = self.p
        ps = []
        m = self.m
        bi = util.tobin(0,m)
        for i in range(1,2**self.m):
            bj = util.tobin(i,m)
            ps = ps + [util.get_string(bi,bj)]
        return ",".join(ps)

    def to_string(self):
        p = self.p
        ps = []
        m = self.m
        bi = util.tobin(0,m)
        for i in range(1,2**self.m):
            bj = util.tobin(i,m)
            ps = ps + [util.get_string(bi,bj)]
        return ps
    
    def properties(self):
        s = f'x family'
        s = s + '\nQubits: ' + str(self.m)
        s = s + '\nCoefficients: ' + str(self.get_coefficients())
        return s
    
    def get_generating_matrix(self):
        return False
    
class zfamily(family):
    def __init__(self,m):
        self.p = False
        self.q = False
        self.m = m
        self.coefficients = "unassigned"
        self.signs = [1]*(2**self.m)
        
    def get_permutation(self):
        return False
    
    def apply_to_circuit(self,circuit):
        return circuit
    
    def __str__(self):
        p = self.p
        ps = []
        m = self.m
        bj = util.tobin(0,m)
        for i in range(0,2**self.m):
            bi = util.tobin(i,m)
            ps = ps + [util.get_string(bi,bj)]
        return ",".join(ps)

    def to_string(self):
        p = self.p
        ps = []
        m = self.m
        bj = util.tobin(0,m)
        for i in range(0,2**self.m):
            bi = util.tobin(i,m)
            ps = ps + [util.get_string(bi,bj)]
        return ps
    
    def properties(self):
        s = f'z family'
        s = s + '\nQubits: ' + str(self.m)
        s = s + '\nCoefficients: ' + str(self.get_coefficients())
        return s
    
    def get_generating_matrix(self):
        return False

    def get_diagonalizing_strings(self):
        return ["".join(["I"]*self.m)]



#Circuit represents a quantum_circuit object from qiskit. This method implements a rotation on this circuit.
#This rotation is the type exp(i pi/4  (Z_j)). This list l contains the qubits which are sigma_z. For example:
#if l = [1,4,5], then Z_j = IZIIZZ (since it has a Z in qubits 1 4 and 5 but not 0 2 and 3)
def apply_e_z(circuit,l):
    for i in range(1,len(l)):           #Start by applying a controlled not gate between the first z qubit
        circuit.cx(l[i],l[0])           #and every other qubit
    
    circuit.sdg(l[0])                     #Next, apply an Sdg gate to the first qubit
    
    for i in range(1,len(l)):           #Finally apply a second set of controlled not gates
        circuit.cx(l[i],l[0])

#Utility method for apply_e_x. The goal is to return a list of the qubits which need to be acted on.
def get_change_list(index,m):
    s = list(bin(index)[2:])        #get binary vector for the index
    l=[]
    first = False
    for i in range(len(s)):
        if(s[len(s)-i-1] == '1'):
            l = l + [i]             #If the ith qubit should be changed, add i to the list l
    return l

#Circuit represents a quantum_circuit object from qiskit. This method implements a rotation on this circuit.
#This rotation is the type exp(i pi/4  (X_index)) where index is an int, and X_index is the member of the X family
#which is represented by this vector. X_5 = XIX etc...
def apply_e_x(circuit,index,m): 
    l = get_change_list(index,m)        #Get the qubits which need to be acted on
    for i in l:
        circuit.u(pi/2,0,0,i)            #Apply y rotation to each of these qubits
    apply_e_z(circuit,l)                #Apply e^(i pi/4 Z_i) (This step involves entanglement)
    for i in l:
        circuit.u(-pi/2,0,0,i)             #Apply y rotation again

#Similar to the previous method apply_e_x, this method is used for repeated applications of this method.
#The rotation exp((i pi/4)(IX + XX)) should be U_y (exp(i pi/4) (IZ + ZZ)) U_y instead of the longer method
#U_y exp(pi/4)(IZ) U_y U_y (exp(pi/4)(ZZ)) U_y. This would be extremely inefficient
def apply_e_x_series(circuit,x_list,m):
    change_lists = []
    for x in x_list:
        change_lists = change_lists + [get_change_list(x,m)]  #Get qubits acted on
    
    #This step could be inefficient in other contexts. We Y rotate all qubits, since it is generally necessary for
    #this problem. However, we really should check which qubits need to be rotated if we want to reuse this method
    for i in range(m):
        circuit.u(-pi/2,0,0,i)                                    #U_y
    for l in change_lists:
        apply_e_z(circuit,l)                                    #exp(Z_j) for each j
    for i in range(m):
        circuit.u(pi/2,0,0,i)                                     #U_y


"""
matrix_generator.py
@author: Ben Reggio 

This code generates a m x m companion matrix and finds a symmetric matrix which is similar to that companion matrix in GF(2)

This is done in a few steps.

First, a companion matrix C is found with companion_matrix
Then, a symmetric matrix F is found with friend_matrix() such that FC^T = CF
M = N^T F N, such that M is symmetric and has all 1s on the diagonal with find_N()
Then, M = L L^T for some L, which is found using split()
Now N^T F N = L L^T so F = [(N^T)^-1 L ] [L^T N^-1] = R R^T where R = (N^T)^-1 L
Since F = R R^T, and F C = C^T F, R^-1 C R = A is symmetric and similar to C. R is our result
"""


import galois as gf     #Generates irreducible_poly for us
from psfam.pauli_utils import *
GF2 = gf.GF(2)          #Our vectors and matrices will be in GF(2)

#Generates a companion matrix from an irreducible polynomial of size m + 1. 
def companion_matrix(m):
  field = gf.GF(2**m)                               #initialize Galois field
  a = list(field.irreducible_poly.coeffs[0:m])       #Find irreducible polynomial
  #print(a)
  C=[]
  for i in range(m-1):                              #Populate companion matrix by row
    row = [GF2(0)]*m
    row[i+1] = GF2(1)
    C = C + [row]
  lastrow = a
  #for i in range(m):
  #  lastrow = lastrow + [a[m-i-1]]
  C = C + [lastrow]
  return C

#Generates a vector which is useful for generating friend matrix F:
def get_b(a):
  m=len(a)
  b = [GF2(1)]                              #b_1 = 1
  for i in range(1,m):
    bn = GF2(0)
    for k in range(0,i):
      bn = bn + a[m-i+k]*b[k]               #Formula for b_i = sum a_{m-i-k} b_k
    b = b + [bn]
    #print("b = " + str(b))
  return b

#Generates a symmetric matrix which satisfies FC^T = CF.
def friend_matrix(C):
  m = len(C)
  a = C[m-1]
  b=get_b(a)                        #Uses vector based off of the irreducible polynomial
  firstrow = [GF2(0)]*m
  firstrow[0] = GF2(1)
  F = []
  for i in range(0,m):
    nextrow = [GF2(0)]*m
    nextrow[m-1-i] = GF2(1)         #Second diagonal should be all ones
    for j in range(1,i+1):
      nextrow[m-1-i + j] =  b[j]
    F = F + [nextrow]
  return F

def find_N(f):
  m = len(f)
  flippers = []
  base = -1
  for i in range(m):
    if(f[i][i] == GF2(1)):
      if base == -1:
        base = i
    else:
      flippers = flippers + [i]
  N = []
  #print("Base = " + str(base))
  for i in range(m):
    nextrow = [GF2(0)]*m
    nextrow[i] = GF2(1)
    if(i == base):
      for j in flippers:
        nextrow[j] = GF2(1)
    N = N + [nextrow]
  return N

#For a symmetric matrix with 1s on the diagonal A, gives L where L L^T = A
def split(M):
  n = len(M)
  if(n == 1):
    return [[GF2(1)]]
  for i in range(n):
    M2 = []
    for j in range(n):
      if(j != i):
        nextrow = M[j].copy()
        nextrow.pop(i)
        M2 = M2 + [nextrow]

    if(is_invertible(M2)):

      L = split(M2)
      if(L):

        if(is_invertible(L)):
          L_prime = invert(L)

          b = M[i].copy()
          b.pop(i)

          eta = mtimes(L_prime,b)

          """
          This removed code is used to verify that a_0 = 1
          s = GF2(0)
          for j in range(len(eta)):
            s = s + eta[j]
          if(s == GF2(1)):
            print("Case: ")
            show_GF_mat(L)
            print()
            show_GF_mat(M)
            print()
            print(eta)
            print(i)
            show_GF_mat(M2)
            print()
            print()
          """

          LR = []
          for j in range(n):
            nextrow = []
            if (j<i):
              nextrow = L[j][0:i] + [GF2(0)] + L[j][i:n-1]
            if(j == i):
              nextrow = eta[0:i] + [GF2(1)] + eta[i:n-1]
            if (j>i):
              nextrow = L[j-1][0:i] + [GF2(0)] + L[j-1][i:n-1]
            LR = LR + [nextrow]
          return LR
        else:
            show_GF_mat(L)
            print("ERROR L NOT INVERTIBLE")
  return False


#Returns a symmetric matrix which is similar to a companion matrix of size m
def get_generating_matrix(m):
    C = companion_matrix(m)

    #print("C = ")
    #for row in C:
    #    print(row)

    F = friend_matrix(C)

    #print("F = ")
    #for row in F:
    #    print(row)

    N = find_N(F)

    #print("N = ")
    #for row in N:
    #    print(row)


    M = mtimes(transpose(N),mtimes(F,N))

    #print("M = ")
    #for row in M:
    #    print(row)

    L = split(M)

    #print("L = ")
    #for row in L:
    #    print(row)

    R=mtimes(transpose(invert(N)),L)
    #NF = mtimes(R,transpose(R))

    #print("NF = ")
    #for row in NF:
    #    print(row)

    A=mtimes(invert(R),mtimes(C,R))
    for i in range(m):
        for j in range(m):
            if(A[i][j] == GF2(1)):
                A[i][j] = 1
            else:
                A[i][j] = 0
    return A


#This method assists in testing. It checks whether a given matrix will work as a generating matrix for a family
def check(A):
  m = len(A)
  N = 2**m
  one = [0]*(m-1) + [1]
  v=one.copy()
  hits = [False]*N
  hits[to_int(v)-1] = True
  for i in range(N-2):
    r = mat_times_vec(A,v)
    r_int = to_int(r)
    if(hits[r_int-1]):
      return False
    else:
      hits[r_int-1] = True
    v = r.copy()
  r = mat_times_vec(A,v)
  return r == one


#TESTING
"""
C = companion_matrix(3)
print("C=")
for row in C:
    print(row)
    
F = friend_matrix(C)
print("F = ")

for row in F:
    print(row)
    
A=get_generating_matrix(5)

for row in A:
    print(row)
for i in range(3,20):
    M = get_generating_matrix(i)
    print(check(M))
"""


import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "psfam",
    version = "0.0.12",
    author = "Ben Reggio",
    author_email = "benjreggio@gmail.com",
    description = ("First attempt at putting psfam in a build."),
    license = "MIT",
    packages=['psfam'],
    long_description=read('README.md'),
    long_description_content_type = "text/markdown",
    install_requires = ['scikit-build','markdown','numpy','qiskit','galois','scipy','functools','itertools','operator'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
)